'JPEG Viewer by Dmitry Brant
'
'QTopia -- http://dmitrybrant.hypermart.net
'dmitrybrant24@hotmail.com
'
'This program is freeware. Use freely, but give me credit where credit
'is due. This is the first and only JPG viewer for QBasic that I have seen.
'Feel free to make improvements, additions, optimizations, or whatever.
'If you do, please tell me or show me what you have done.
'This program is also provided on an "as-is" basis.
'I can't be held responsible for the mischevious deeds you may perform
'with the aid of this program.

'IMPORTANT: For this program to run properly, you MUST start QuickBasic
'with the following command line:

'QB /L LIBVESA.QLB

'This QLB file should be distributed along with this program.
'The QLB file already contains the standard QB.QLB file, so don't worry.
'Also, the include file LIBVESA.BI must be in the same directory as
'this program.

'NOTE: This specific edition of the Viewer is designed to work in
'HiColor or TrueColor screen modes. To use these screen modes, your
'VGA card must have VESA compatibility. If you find that this program
'does not work with your VGA card, try installing a VESA BIOS extension
'like UNIVBE.EXE.

'Read the README.TXT file for more information.


'These are the only SUBs that have to be called to display the image.
'There are three different subs for three different
'mode types (15-bit, 16-bit, and 24-bit).
'For example, to display an image in a 16-bit mode, you must use PutJPG16.

'For 15-bit (HiColor) modes:
DECLARE SUB PutJPG15 (file$, x0%, y0%)
    'file$ = filename of the JPG image
    'x0, y0 = coordinates where to start drawing the image

'For 16-bit (HiColor) modes:
DECLARE SUB PutJPG16 (file$, x0%, y0%)
    'file$ = filename of the JPG image
    'x0, y0 = coordinates where to start drawing the image

'For 24-bit (TrueColor) modes:
DECLARE SUB PutJPG24 (file$, x0%, y0%)
    'file$ = filename of the JPG image
    'x0, y0 = coordinates where to start drawing the image


'These other functions are called from within the PutJPG sub.
'They are useless by themselves.
DECLARE SUB QuantReorder (inArr%(), num%)
DECLARE FUNCTION Decode% (inArray() AS ANY, hnum%)
DECLARE FUNCTION ReceiveBits% (cat%)
DECLARE FUNCTION NextBit% ()
DECLARE SUB ToRGB (y0%, cb0%, cr0%, r0%, g0%, b0%)
DECLARE SUB GetBlock (vector%(), HuffDC() AS ANY, HuffDCNum%, HuffAC() AS ANY, HuffACNum%, Quant%(), QuantNum%, dcCoef%)
DECLARE FUNCTION GetHuffTables% ()
DECLARE FUNCTION GetQuantTables% (inArray() AS INTEGER)
DECLARE FUNCTION GetAPP0% ()
DECLARE FUNCTION GetSOS% ()
DECLARE FUNCTION GetWord& ()
DECLARE FUNCTION GetByte% ()
DECLARE FUNCTION GetSOF% ()
DECLARE FUNCTION GetImageAttr% ()
DECLARE FUNCTION Bin$ (De&, bits%)
DECLARE SUB IDCT (array() AS INTEGER)
DECLARE SUB waitkey ()
DEFINT A-Z
'$INCLUDE: 'qb.bi'
'$INCLUDE: 'libvesa.bi'

TYPE JpegType               'some type definitions (for better coherence)
    Rows AS INTEGER             'image height
    Cols AS INTEGER             'image width
    SamplesY AS INTEGER         'sampling ratios
    SamplesCbCr AS INTEGER
    QuantTableY AS INTEGER      'quantization table numbers
    QuantTableCbCr AS INTEGER
    HuffDCTableY AS INTEGER     'huffman table numbers
    HuffDCTableCbCr AS INTEGER
    HuffACTableY AS INTEGER
    HuffACTableCbCr AS INTEGER
END TYPE

TYPE HuffmanEntry                 'a type for huffman tables
    Index AS LONG
    Code AS INTEGER
    Length AS INTEGER
END TYPE
                                  'a few global variables
COMMON SHARED a$
COMMON SHARED curByte, curBits, jfile, EOI, DCTables, ACTables, QTables
COMMON SHARED Xres, yres, yOffset, curBank, winGran, oldMode, ppSeg, ppOfs, bpp
DIM SHARED Regs AS RegTypeX
DIM SHARED image AS JpegType      'Global image attributes
DIM SHARED HuffmanDC(0 TO 1, 0 TO 255) AS HuffmanEntry '2 Huffman DC tables
DIM SHARED HuffmanAC(0 TO 1, 0 TO 255) AS HuffmanEntry '2 Huffman AC tables
CONST pi = 3.1415926535#          'needed for the IDCT


'--------- A demonstration ----------
'
'The program will attempt to go into mode 640 x 480 x 64k (HiColor)
'and then display a sample image.

CLS
INPUT "Enter a valid .JPG image file name> ", f$

SCREEN 13
IF DetectVESA THEN
  pagesAvailable = SetSVGAMode(&H111)  'mode 0x111 = 640x480x64k
  IF pagesAvailable = 0 THEN
    PRINT "Could not set HiColor mode."
  END IF
ELSE
  CLS
  PRINT "VESA Compatibility not detected."
  PRINT "Recommend installing UNIVBE.EXE."
  SYSTEM
END IF

'We use the PutJPG16 function because we are in a 16-bit mode.

PutJPG16 f$, 50, 50       'draw the image at offset 50, 50 (for example)

BEEP
waitkey
SCREEN 0: WIDTH 80, 25

SYSTEM

'--------- End of Demo ----------



Zig2:        'Zigzag patterns for reordering quantization tables and vectors
DATA 0,0     
ZigzagPositions:
DATA 0,1,1,0
DATA 2,0,1,1,0,2
DATA 0,3,1,2,2,1,3,0
DATA 4,0,3,1,2,2,1,3,0,4
DATA 0,5,1,4,2,3,3,2,4,1,5,0
DATA 6,0,5,1,4,2,3,3,2,4,1,5,0,6
DATA 0,7,1,6,2,5,3,4,4,3,5,2,6,1,7,0
DATA 7,1,6,2,5,3,4,4,3,5,2,6,1,7
DATA 2,7,3,6,4,5,5,4,6,3,7,2
DATA 7,3,6,4,5,5,4,6,3,7
DATA 4,7,5,6,6,5,7,4
DATA 7,5,6,6,5,7
DATA 6,7,7,6
DATA 7,7

FUNCTION Decode (inArray() AS HuffmanEntry, hnum)

n = GetByte
IF n = 255 THEN
    n1 = GetByte
    IF n1 >= &HD0 AND n1 <= &HD7 THEN
        n2 = 2 ^ curBits - 1
        SEEK #jfile, SEEK(jfile) - 2
        IF curByte AND n2 = n2 THEN    'if the remaining bits are 1
            EOI = 1
            Decode = 0
            EXIT FUNCTION
        END IF
    ELSE
        SEEK #jfile, SEEK(jfile) - 2
    END IF
ELSE
    SEEK #jfile, SEEK(jfile) - 1
END IF

curVal& = 0
MatchFound = -1
FOR l = 1 TO 16    'cycle through 16 possible Huffman lengths
    curVal& = curVal& * 2 + NextBit
    IF EOI THEN EXIT FUNCTION

    FOR i = 0 TO 255              'look for a match in the Huffman table
        IF inArray(hnum, i).Length > l THEN EXIT FOR
        IF inArray(hnum, i).Length = l THEN
            IF inArray(hnum, i).Index = curVal& THEN MatchFound = i: EXIT FOR
        END IF
    NEXT i
    IF MatchFound > -1 THEN EXIT FOR
NEXT l

IF MatchFound = -1 THEN BEEP: Decode = -1: EXIT FUNCTION

Decode = inArray(hnum, MatchFound).Code  'return the appropriate code
END FUNCTION

' Detect to see if a VESA compatible graphics card is present.
FUNCTION DetectVESA

DIM VGAInfo AS VGAInfoBlock

Regs.ax = &H4F00
Regs.es = VARSEG(VGAInfo)
Regs.di = VARPTR(VGAInfo)
CALL INTERRUPTX(&H10, Regs, Regs)

IF Regs.ax = &H4F THEN
  DetectVESA = 1
  PRINT VGAInfo.VESASignature; RTRIM$(STR$(VGAInfo.VESAVersion \ 256)); "."; LTRIM$(STR$(VGAInfo.VESAVersion AND 255))
ELSE
  DetectVESA = 0
  PRINT "Error - cannot detect VESA."
END IF

END FUNCTION

FUNCTION GetAPP0
'Obtains the JFIF APP0 marker
'Enable the PRINTs if you want to see the results

'PRINT "JFIF APP0 Marker"
temp1& = GetWord
'PRINT "  Length:";temp1
'PRINT "  JFIF Header: ";
a$ = "     "
GET #1, , a$
'IF a$ = "JFIF" + CHR$(0) THEN PRINT "Check." ELSE PRINT "Failed."
a$ = " "
temp0 = GetByte
'PRINT "  Major Revision #";temp0
temp0 = GetByte
'PRINT "  Minor Revision #"; temp0
temp0 = GetByte
SELECT CASE temp0
    CASE 0
        'PRINT "  x/y density specify aspect ratio"
    CASE 1
        'PRINT "  x/y density are dots/inch"
    CASE 2
        'PRINT "  x/y density are dots/cm"
END SELECT
temp0 = GetByte
'PRINT "  X-Density:"; temp0
temp0 = GetByte
'PRINT "  Y-Density:"; temp0
temp0 = GetByte
'PRINT "  Thumbnail Width:"; temp0
temp1 = GetByte
'PRINT "  Thumbnail Height:"; temp1
'PRINT "  Bytes for Thumbnail:"; 3 * temp0 * temp1
GetAPP0 = 1
END FUNCTION

SUB GetBlock (vector(), HuffDC() AS HuffmanEntry, HuffDCNum, HuffAC() AS HuffmanEntry, HuffACNum, Quant(), QuantNum, dcCoef)
EOI = 0

temp0 = Decode(HuffDC(), HuffDCNum)   'Get the DC coefficient
IF EOI THEN d = 0
diff = ReceiveBits(temp0)
dcCoef = dcCoef + diff
vector(0, 0) = dcCoef

xpos = 0: ypos = 0
RESTORE ZigzagPositions
ACcount = 1
DO
    d = Decode(HuffAC(), HuffACNum)
    IF EOI THEN d = 0


    zeros = d \ 16
    bits = d AND 15
    bitVal = ReceiveBits(bits)

    IF zeros = 0 AND bits = 0 THEN   'EOB Encountered
        FOR j = ACcount TO 63
            READ xpos, ypos
            vector(xpos, ypos) = 0
        NEXT j
        EXIT DO
    ELSEIF zeros = 15 AND bits = 0 THEN  'ZRL encountered
        FOR j = ACcount TO ACcount + 15
            READ xpos, ypos
            vector(xpos, ypos) = 0
        NEXT j
        ACcount = ACcount + 16
    ELSE
        FOR j = 1 TO zeros
            READ xpos, ypos
            vector(xpos, ypos) = 0
            ACcount = ACcount + 1
            IF ACcount >= 64 THEN EXIT DO
        NEXT j
        READ xpos, ypos
        vector(xpos, ypos) = bitVal
        ACcount = ACcount + 1
    END IF
    IF ACcount >= 64 THEN EXIT DO
LOOP

FOR i = 0 TO 7
    FOR j = 0 TO 7
        vector(i, j) = vector(i, j) * Quant(QuantNum, i, j)
    NEXT j
NEXT i

IDCT vector()

FOR i = 0 TO 7
    FOR j = 0 TO 7
        vector(i, j) = vector(i, j) + 128
    NEXT j
NEXT i
END SUB

FUNCTION GetByte
GET #jfile, , a$
GetByte = ASC(a$)
END FUNCTION

FUNCTION GetHuffTables
DIM HuffAmount(1 TO 16)
l0 = GetWord

c0 = 2
DO
temp0 = GetByte: c0 = c0 + 1
t0 = (temp0 AND 16) \ 16
temp0 = temp0 AND 15
SELECT CASE t0
    CASE 0        'DC Table
        total = 0
        FOR i = 1 TO 16
            temp1 = GetByte: c0 = c0 + 1
            total = total + temp1
            HuffAmount(i) = temp1
        NEXT i
        FOR i = 0 TO total - 1
            HuffmanDC(temp0, i).Code = GetByte: c0 = c0 + 1
        NEXT i
       
        curNum& = 0
        curIndex = -1
        FOR i = 1 TO 16
            FOR j = 1 TO HuffAmount(i)
                curIndex = curIndex + 1
                HuffmanDC(temp0, curIndex).Index = curNum&
                HuffmanDC(temp0, curIndex).Length = i
                curNum& = curNum& + 1
            NEXT j
            curNum& = curNum& * 2
        NEXT i

        'PRINT "Huffman DC Table"; temp0; "Retrieved."
        DCTables = DCTables + 1
    CASE 1
        total = 0
        FOR i = 1 TO 16
            temp1 = GetByte: c0 = c0 + 1
            total = total + temp1
            HuffAmount(i) = temp1
        NEXT i
        FOR i = 0 TO total - 1
            HuffmanAC(temp0, i).Code = GetByte: c0 = c0 + 1
        NEXT i

        curNum& = 0
        curIndex = -1
        FOR i = 1 TO 16
            FOR j = 1 TO HuffAmount(i)
                curIndex = curIndex + 1
                HuffmanAC(temp0, curIndex).Index = curNum&
                HuffmanAC(temp0, curIndex).Length = i
                curNum& = curNum& + 1
            NEXT j
            curNum& = curNum& * 2
        NEXT i

        'PRINT "Huffman AC Table"; temp0; "Retrieved."
        ACTables = ACTables + 1
END SELECT
LOOP UNTIL c0 >= l0

GetHuffTables = 1
END FUNCTION

FUNCTION GetImageAttr
'PRINT "Start of Frame 0:"
temp4& = GetWord
'PRINT "  Length:";temp4&
temp0 = GetByte
IF temp0 <> 8 THEN GetImageAttr = 0: EXIT FUNCTION   'we do not support 12 or 16-bit samples
'PRINT "  Data Precision in bits/sample:"; temp0
image.Rows = GetWord
'PRINT "  Image Height:"; image.Rows
image.Cols = GetWord
'PRINT "  Image Width:"; image.Cols
temp0 = GetByte
SELECT CASE temp0
    CASE 1
        'PRINT "  Grayscale image."
    CASE 3
        'PRINT "  Color image (YCbCr)"
    CASE ELSE
        GetImageAttr = 0: EXIT FUNCTION     'unsupported
END SELECT
FOR i = 1 TO temp0
    id = GetByte
    SELECT CASE id
        CASE 1
            'PRINT "    Component Y:"
            temp1 = GetByte
            image.SamplesY = (temp1 AND 15) * (temp1 \ 16)
            'PRINT "    Number of samples:"; image.SamplesY
            image.QuantTableY = GetByte
            'PRINT "    Quantization table to use:"; image.QuantTableY
        CASE 2, 3
            'PRINT "    Component Cb/Cr:"
            temp1 = GetByte
            image.SamplesCbCr = (temp1 AND 15) * (temp1 \ 16)
            'PRINT "    Number of samples:"; image.SamplesCbCr
            image.QuantTableCbCr = GetByte
            'PRINT "    Quantization table to use:"; image.QuantTableCbCr
    END SELECT
NEXT i
'waitkey
GetImageAttr = 1
END FUNCTION

FUNCTION GetQuantTables (inArray() AS INTEGER)
l0 = GetWord

c0 = 2
DO
temp0 = GetByte: c0 = c0 + 1
IF (temp0 AND &HF0) <> 0 THEN
    GetQuantTables = 0: EXIT FUNCTION  'we don't support 16-bit tables
END IF
temp0 = temp0 AND 15
'PRINT "  Quantization Table Number:"; temp0
'PRINT "  Table:"
FOR i = 0 TO 7
    FOR k = 0 TO 7
        temp1 = GetByte: c0 = c0 + 1
        'PRINT temp1;
        inArray(temp0, i, k) = temp1
    NEXT k
    'PRINT
NEXT i
'waitkey
QTables = QTables + 1
QuantReorder inArray(), temp0
LOOP UNTIL c0 >= l0

GetQuantTables = 1: EXIT FUNCTION
END FUNCTION

FUNCTION GetSOF
a$ = "  "
SEEK #jfile, 1
GET #jfile, , a$
IF a$ = CHR$(255) + CHR$(&HD8) THEN d = 1 ELSE d = 0
a$ = " "
GetSOF = d
END FUNCTION

FUNCTION GetSOS
'PRINT "Start Of Scan"
temp4& = GetWord
'PRINT "  Length:"; temp4&
temp0 = GetByte
IF temp0 <> 1 AND temp0 <> 3 THEN GetSOS = 0: EXIT FUNCTION
'PRINT "  Number of Components in Scan:"; temp0
FOR i = 1 TO temp0
    'PRINT "  Component ID: ";
    temp1 = GetByte
    SELECT CASE temp1
        CASE 1
            'PRINT "Y"
            temp2 = GetByte
            'PRINT "  Huffman Table to Use:"
            image.HuffACTableY = temp2 AND &HF
            'PRINT "    AC Table #"; image.HuffACTableY
            image.HuffDCTableY = temp2 \ 16
            'PRINT "    DC Table #"; image.HuffDCTableY
        CASE 2
            'PRINT "Cb"
            temp2 = GetByte
            'PRINT "  Huffman Table to Use:"
            image.HuffACTableCbCr = temp2 AND &HF
            'PRINT "    AC Table #"; image.HuffACTableCbCr
            image.HuffDCTableCbCr = temp2 \ 16
            'PRINT "    DC Table #"; image.HuffDCTableCbCr
        CASE 3
            'PRINT "Cr"
            temp2 = GetByte
            'PRINT "  Huffman Table to Use:"
            image.HuffACTableCbCr = temp2 AND &HF
            'PRINT "    AC Table #"; image.HuffACTableCbCr
            image.HuffDCTableCbCr = (temp2 AND &HF0) \ 16
            'PRINT "    DC Table #"; image.HuffDCTableCbCr
        CASE ELSE
            GetSOS = 0: EXIT FUNCTION
    END SELECT
NEXT i
a$ = "   "
GET #1, , a$
a$ = " "
'waitkey
GetSOS = 1
END FUNCTION

' Returns the current video mode.
FUNCTION GetVideoMode

Regs.ax = &H4F03
CALL INTERRUPTX(&H10, Regs, Regs)
GetVideoMode = Regs.bx

END FUNCTION

FUNCTION GetWord&
GET #jfile, , a$
l0& = CLNG(ASC(a$)) * 256
GET #jfile, , a$
l0& = l0& + ASC(a$)
GetWord& = l0&
END FUNCTION

SUB IDCT (array() AS INTEGER)    'array must be 8x8
DIM array2(0 TO 7, 0 TO 7)

FOR x = 0 TO 7
    FOR y = 0 TO 7
        sum# = 0
        temp# = 0
        FOR u = 0 TO 7
            FOR v = 0 TO 7
                temp# = array(u, v) * COS((2 * x + 1) * u * pi / 16)
                temp# = temp# * COS((2 * y + 1) * v * pi / 16)
                IF v = 0 AND u = 0 THEN temp# = temp# / 2
                sum# = sum# + temp#
            NEXT v
        NEXT u
        sum# = sum# / 4
        array2(x, y) = sum#
    NEXT y
NEXT x

FOR u = 0 TO 7
    FOR v = 0 TO 7
        array(u, v) = array2(u, v)
    NEXT v
NEXT u
END SUB

FUNCTION NextBit

t0 = 2 ^ curBits
v0 = ABS((curByte AND t0) <> 0)

curBits = curBits - 1
IF curBits < 0 THEN
    curBits = 7: curByte = GetByte
    IF curByte = 255 THEN
        b2 = GetByte
        IF b2 = &HD9 THEN EOI = 1: EXIT FUNCTION
    END IF
END IF

NextBit = v0
END FUNCTION

SUB PutJPG15 (file$, x0, y0)
DIM YVector1(0 TO 7, 0 TO 7)              '4 vectors for Y attribute
DIM YVector2(0 TO 7, 0 TO 7)
DIM YVector3(0 TO 7, 0 TO 7)
DIM YVector4(0 TO 7, 0 TO 7)
DIM CbVector(0 TO 7, 0 TO 7)              '1 vector for Cb attribute
DIM CrVector(0 TO 7, 0 TO 7)              '1 vector for Cr attribute
DIM QuantTable(0 TO 1, 0 TO 7, 0 TO 7)    '2 quantization tables (Y, CbCr)

jfile = FREEFILE
a$ = " "   'The ideal byte
OPEN file$ FOR BINARY AS #jfile
IF LOF(jfile) = 0 THEN CLOSE #jfile: KILL file$: CLS : PRINT "File does not exist.": EXIT SUB
IF GetSOF = 0 THEN CLS : PRINT "Not a valid JPEG/JFIF file.": CLOSE #jfile: EXIT SUB

QTables = 0           'Initialize some checkpoint variables
ACTables = 0
DCTables = 0
Restart& = 0

DO                            'Primary control loop for markers

    IF GetByte = 255 THEN     'Marker Found
        d = GetByte
        SELECT CASE d         'which one is it?
            CASE &HC0         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC1         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC9         'SOF9
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HC4         'DHT
                IF ACTables < 2 OR DCTables < 2 THEN
                    IF GetHuffTables = 0 THEN PRINT "Error getting Huffman tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HCC         'DAC
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HD8         'SOI
                'PRINT "Start Of Image"
            CASE &HD9         'EOI
                'PRINT "End Of Image"
            CASE &HDA         'SOS
                IF GetSOS = 0 THEN PRINT "Error getting SOS marker.": CLOSE #jfile: EXIT SUB
                IF DCTables = 2 AND ACTables = 2 AND QTables = 2 THEN
                    EOI = 0
                    EXIT DO                 'Go on to secondary control loop
                ELSE
                    PRINT "Unsupported file format.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDB         'DQT
                IF QTables < 2 THEN
                    IF GetQuantTables(QuantTable()) = 0 THEN PRINT "Error getting quantization tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDD         'DRI
                'PRINT "Define Restart Interval"
                Restart& = GetWord
                'PRINT "  Restart Interval (in MCU blocks):"; Restart&
            CASE &HE0         'APP0
                IF GetAPP0 = 0 THEN PRINT "Error getting APP0 marker.": CLOSE #jfile: EXIT SUB
            CASE &HFE         'COM
                'PRINT "Comment"
                a$ = SPACE$(GetWord - 2)
                GET #1, , a$
                'PRINT "  " + a$
                a$ = " "
        END SELECT

    END IF
    r$ = INKEY$
    IF r$ = CHR$(27) THEN CLOSE #jfile: EXIT SUB
LOOP UNTIL EOF(1)

xpos = 0: ypos = 0
dcY = 0
dcCb = 0
dcCr = 0
xindex = 0: yindex = 0
mcu = 0

curBits = 7
curByte = GetByte

r = g = b = 0
DEF SEG = &HA000

SELECT CASE image.SamplesY
CASE 4

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector3(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector4(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 0 TO 7
            y = YVector3(i - 8, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 8 TO 15
            y = YVector4(i - 8, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 16: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 2

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 1

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            cb = CbVector(i, j)
            cr = CrVector(i, j)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put15 xj + x0, yi + y0, r \ 8, g \ 8, b \ 8
        NEXT j
    NEXT i
 
    xindex = xindex + 8
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

END SELECT

CLOSE #jfile

END SUB

SUB PutJPG16 (file$, x0, y0)
DIM YVector1(0 TO 7, 0 TO 7)              '4 vectors for Y attribute
DIM YVector2(0 TO 7, 0 TO 7)
DIM YVector3(0 TO 7, 0 TO 7)
DIM YVector4(0 TO 7, 0 TO 7)
DIM CbVector(0 TO 7, 0 TO 7)              '1 vector for Cb attribute
DIM CrVector(0 TO 7, 0 TO 7)              '1 vector for Cr attribute
DIM QuantTable(0 TO 1, 0 TO 7, 0 TO 7)    '2 quantization tables (Y, CbCr)

jfile = FREEFILE
a$ = " "   'The ideal byte
OPEN file$ FOR BINARY AS #jfile
IF LOF(jfile) = 0 THEN CLOSE #jfile: KILL file$: CLS : PRINT "File does not exist.": EXIT SUB
IF GetSOF = 0 THEN CLS : PRINT "Not a valid JPEG/JFIF file.": CLOSE #jfile: EXIT SUB

QTables = 0           'Initialize some checkpoint variables
ACTables = 0
DCTables = 0
Restart& = 0

DO                            'Primary control loop for markers

    IF GetByte = 255 THEN     'Marker Found
        d = GetByte
        SELECT CASE d         'which one is it?
            CASE &HC0         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC1         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC9         'SOF9
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HC4         'DHT
                IF ACTables < 2 OR DCTables < 2 THEN
                    IF GetHuffTables = 0 THEN PRINT "Error getting Huffman tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HCC         'DAC
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HD8         'SOI
                'PRINT "Start Of Image"
            CASE &HD9         'EOI
                'PRINT "End Of Image"
            CASE &HDA         'SOS
                IF GetSOS = 0 THEN PRINT "Error getting SOS marker.": CLOSE #jfile: EXIT SUB
                IF DCTables = 2 AND ACTables = 2 AND QTables = 2 THEN
                    EOI = 0
                    EXIT DO                 'Go on to secondary control loop
                ELSE
                    PRINT "Unsupported file format.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDB         'DQT
                IF QTables < 2 THEN
                    IF GetQuantTables(QuantTable()) = 0 THEN PRINT "Error getting quantization tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDD         'DRI
                'PRINT "Define Restart Interval"
                Restart& = GetWord
                'PRINT "  Restart Interval (in MCU blocks):"; Restart&
            CASE &HE0         'APP0
                IF GetAPP0 = 0 THEN PRINT "Error getting APP0 marker.": CLOSE #jfile: EXIT SUB
            CASE &HFE         'COM
                'PRINT "Comment"
                a$ = SPACE$(GetWord - 2)
                GET #1, , a$
                'PRINT "  " + a$
                a$ = " "
        END SELECT
 
    END IF
    r$ = INKEY$
    IF r$ = CHR$(27) THEN CLOSE #jfile: EXIT SUB
LOOP UNTIL EOF(1)

xpos = 0: ypos = 0
dcY = 0
dcCb = 0
dcCr = 0
xindex = 0: yindex = 0
mcu = 0

curBits = 7
curByte = GetByte

r = g = b = 0
DEF SEG = &HA000

SELECT CASE image.SamplesY
CASE 4

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector3(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector4(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr
 
    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 0 TO 7
            y = YVector3(i - 8, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 8 TO 15
            y = YVector4(i - 8, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 16: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 2

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 1

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr
 
    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            cb = CbVector(i, j)
            cr = CrVector(i, j)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put16 xj + x0, yi + y0, r \ 8, g \ 4, b \ 8
        NEXT j
    NEXT i
  
    xindex = xindex + 8
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

END SELECT

CLOSE #jfile
END SUB

SUB PutJPG24 (file$, x0, y0)
DIM YVector1(0 TO 7, 0 TO 7)              '4 vectors for Y attribute
DIM YVector2(0 TO 7, 0 TO 7)
DIM YVector3(0 TO 7, 0 TO 7)
DIM YVector4(0 TO 7, 0 TO 7)
DIM CbVector(0 TO 7, 0 TO 7)              '1 vector for Cb attribute
DIM CrVector(0 TO 7, 0 TO 7)              '1 vector for Cr attribute
DIM QuantTable(0 TO 1, 0 TO 7, 0 TO 7)    '2 quantization tables (Y, CbCr)

jfile = FREEFILE
a$ = " "   'The ideal byte
OPEN file$ FOR BINARY AS #jfile
IF LOF(jfile) = 0 THEN CLOSE #jfile: KILL file$: CLS : PRINT "File does not exist.": EXIT SUB
IF GetSOF = 0 THEN CLS : PRINT "Not a valid JPEG/JFIF file.": CLOSE #jfile: EXIT SUB

QTables = 0           'Initialize some checkpoint variables
ACTables = 0
DCTables = 0
Restart& = 0

DO                            'Primary control loop for markers

    IF GetByte = 255 THEN     'Marker Found
        d = GetByte
        SELECT CASE d         'which one is it?
            CASE &HC0         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC1         'SOF0
                IF GetImageAttr = 0 THEN PRINT "Error getting Start Of Frame 0 Marker.": CLOSE #jfile: EXIT SUB
            CASE &HC9         'SOF9
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HC4         'DHT
                IF ACTables < 2 OR DCTables < 2 THEN
                    IF GetHuffTables = 0 THEN PRINT "Error getting Huffman tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HCC         'DAC
                PRINT "Arithmetic Coding Not Supported.": CLOSE #jfile: EXIT SUB
            CASE &HD8         'SOI
                'PRINT "Start Of Image"
            CASE &HD9         'EOI
                'PRINT "End Of Image"
            CASE &HDA         'SOS
                IF GetSOS = 0 THEN PRINT "Error getting SOS marker.": CLOSE #jfile: EXIT SUB
                IF DCTables = 2 AND ACTables = 2 AND QTables = 2 THEN
                    EOI = 0
                    EXIT DO                 'Go on to secondary control loop
                ELSE
                    PRINT "Unsupported file format.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDB         'DQT
                IF QTables < 2 THEN
                    IF GetQuantTables(QuantTable()) = 0 THEN PRINT "Error getting quantization tables.": CLOSE #jfile: EXIT SUB
                END IF
            CASE &HDD         'DRI
                'PRINT "Define Restart Interval"
                Restart& = GetWord
                'PRINT "  Restart Interval (in MCU blocks):"; Restart&
            CASE &HE0         'APP0
                IF GetAPP0 = 0 THEN PRINT "Error getting APP0 marker.": CLOSE #jfile: EXIT SUB
            CASE &HFE         'COM
                'PRINT "Comment"
                a$ = SPACE$(GetWord - 2)
                GET #1, , a$
                'PRINT "  " + a$
                a$ = " "
        END SELECT

    END IF
    r$ = INKEY$
    IF r$ = CHR$(27) THEN CLOSE #jfile: EXIT SUB
LOOP UNTIL EOF(1)

xpos = 0: ypos = 0
dcY = 0
dcCb = 0
dcCr = 0
xindex = 0: yindex = 0
mcu = 0

curBits = 7
curByte = GetByte

r = g = b = 0
DEF SEG = &HA000

SELECT CASE image.SamplesY
CASE 4

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector3(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector4(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 0 TO 7
            y = YVector3(i - 8, j)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    FOR i = 8 TO 15
        FOR j = 8 TO 15
            y = YVector4(i - 8, j - 8)
            i2 = i \ 2
            j2 = j \ 2
            cb = CbVector(i2, j2)
            cr = CrVector(i2, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 16: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 2

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock YVector2(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    FOR i = 0 TO 7
        FOR j = 8 TO 15
            y = YVector2(i, j - 8)
            j2 = j \ 2
            cb = CbVector(i, j2)
            cr = CrVector(i, j2)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i

    xindex = xindex + 16
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

CASE 1

DO
    GetBlock YVector1(), HuffmanDC(), image.HuffDCTableY, HuffmanAC(), image.HuffACTableY, QuantTable(), image.QuantTableY, dcY
    GetBlock CbVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCb
    GetBlock CrVector(), HuffmanDC(), image.HuffDCTableCbCr, HuffmanAC(), image.HuffACTableCbCr, QuantTable(), image.QuantTableCbCr, dcCr

    FOR i = 0 TO 7
        FOR j = 0 TO 7
            y = YVector1(i, j)
            cb = CbVector(i, j)
            cr = CrVector(i, j)
            ToRGB y, cb, cr, r, g, b
            xj = xindex + j
            yi = yindex + i
            IF xj < image.Cols AND yi < image.Rows THEN Put24 xj + x0, yi + y0, b, g, r
        NEXT j
    NEXT i
 
    xindex = xindex + 8
    IF xindex >= image.Cols THEN xindex = 0: yindex = yindex + 8: mcu = 1
    IF yindex >= image.Rows THEN EXIT DO

    IF mcu = 1 AND Restart& <> 0 THEN 'execute the restart interval
        mcu = 0
        curByte = GetByte
        curByte = GetByte
        curByte = GetByte
        curBits = 7
        dcY = 0
        dcCb = 0
        dcCr = 0
    END IF

    r$ = INKEY$
    IF r$ = CHR$(27) THEN EXIT DO
LOOP

END SELECT

CLOSE #jfile

END SUB

SUB QuantReorder (inArr(), num)
DIM arr2(0 TO 7, 0 TO 7)
RESTORE Zig2
xp = 0: yp = 0
FOR i = 0 TO 7
    FOR j = 0 TO 7
        READ xp, yp
        arr2(xp, yp) = inArr(num, i, j)
    NEXT j
NEXT i
FOR i = 0 TO 7
    FOR j = 0 TO 7
        inArr(num, i, j) = arr2(i, j)
    NEXT j
NEXT i
END SUB

FUNCTION ReceiveBits (cat)
temp0& = 0
FOR i = 1 TO cat
    temp0& = temp0& * 2 + NextBit
NEXT i

IF temp0& >= 2 ^ (cat - 1) THEN
    RealValue = temp0&
ELSE
    RealValue = -(2 ^ cat - 1) + temp0&
END IF

ReceiveBits = RealValue
END FUNCTION

' Sets an SVGA mode.
FUNCTION SetSVGAMode (mode)

oldMode = GetVideoMode                          'Get the current mode.

DIM ModeInfo AS ModeInfoBlock                   'Get info on the desired
Regs.ax = &H4F01                                'SVGA mode.
Regs.cx = mode
Regs.es = VARSEG(ModeInfo)
Regs.di = VARPTR(ModeInfo)
CALL INTERRUPTX(&H10, Regs, Regs)

IF (ModeInfo.ModeAttributes AND 1) = 0 THEN     'Bit 1 = 0 then mode not
  CLS
  PRINT "Error - screen mode not available."    'available.
  SetSVGAMode = 0
  PRINT ModeInfo.BitsPerPixel
  EXIT FUNCTION
END IF
winGran = 64 \ ModeInfo.WinGranularity          'Window granularity adjusted.
Xres = ModeInfo.XResolution                     'Get screen resolution.
yres = ModeInfo.YResolution
IF ModeInfo.WinSize < 64 THEN
  CLS
  PRINT "Error - this program requires at least a 64k memory window."
  SetSVGAMode = 0                               'Window must be > 64k for
  EXIT FUNCTION                                 'these routines.
END IF
bpp = ASC(ModeInfo.BitsPerPixel)

IF bpp = 15 THEN
  IF ASC(ModeInfo.MemoryModel) <> 6 THEN
    CLS
    PRINT "Error - screen mode not available."
    SetSVGAMode = 0
    EXIT FUNCTION
  END IF

  ppSeg = ModeInfo.WinASegment
  DEF SEG = ppSeg
ELSEIF bpp = 16 THEN
  IF ASC(ModeInfo.MemoryModel) <> 6 THEN
    CLS
    PRINT "Error - screen mode not available."
    SetSVGAMode = 0
    EXIT FUNCTION
  END IF

  ppSeg = ModeInfo.WinASegment
  DEF SEG = ppSeg
ELSEIF bpp = 24 THEN
  ppSeg = ModeInfo.WinASegment
  DEF SEG = ppSeg
ELSE
  CLS
  PRINT "Error - screen mode not available."
  PRINT ModeInfo.MemoryModel
  SetSVGAMode = 0
  EXIT FUNCTION
END IF

Regs.ax = &H4F02                                'Set the mode.
Regs.bx = mode
CALL INTERRUPTX(&H10, Regs, Regs)

IF Regs.ax <> &H4F THEN
  CLS
  PRINT "Error - cannot initialize screen mode."
  SetSVGAMode = 0
  EXIT FUNCTION
END IF

curBank = 0
yOffset = 0

SVGAinit winGran, Xres, yres, brr, ppSeg, ASC(ModeInfo.NumberOfImagePages) + 1

SetSVGAMode = (ASC(ModeInfo.NumberOfImagePages) + 1)

END FUNCTION

SUB ToRGB (y0, cb0, cr0, r0, g0, b0)
r0 = y0 + 1.402 * (cr0 - 128)
g0 = y0 - .34414 * (cb0 - 128) - .71414 * (cr0 - 128)
b0 = y0 + 1.772 * (cb0 - 128)

IF r0 > 255 THEN r0 = 255
IF r0 < 0 THEN r0 = 0
IF g0 > 255 THEN g0 = 255
IF g0 < 0 THEN g0 = 0
IF b0 > 255 THEN b0 = 255
IF b0 < 0 THEN b0 = 0
END SUB

SUB waitkey
DO: LOOP UNTIL INKEY$ <> ""
END SUB

